@@ -18,6 +18,7 @@ use std::sync::Arc;
 use chrono::{DateTime, Local};
 use console::Term;
 use crossbeam_utils::thread;
+use indexmap::IndexMap;
 use itertools::Itertools;
 use rayon::prelude::*;
 use serde::*;
@@ -189,6 +190,7 @@ pub struct FileGroup<F> {
 
 /// Controls the type of search by determining the number of replicas
 /// allowed in a group of identical files.
+#[derive(Debug)]
 pub enum Replication {
     /// Looks for under-replicated files with replication factor lower than the specified number.
     /// `Underreplicated(2)` means searching for unique files.
@@ -206,6 +208,7 @@ pub enum Replication {
 /// discarded.
 ///
 /// This is to be configured from the command line parameters set by the user.
+#[derive(Debug)]
 pub struct FileGroupFilter {
     /// The allowed number of replicas in the group.
     pub replication: Replication,
@@ -214,6 +217,8 @@ pub struct FileGroupFilter {
     /// If empty - no additional grouping is performed.
     /// See [`GroupConfig::isolate`].
     pub root_paths: Vec<Path>,
+    /// If set to true, files with the same `FileId` are counted as one
+    pub group_by_id: bool,
 }
 
 impl<F> FileGroup<F> {
@@ -228,7 +233,14 @@ impl<F> FileGroup<F> {
     }
 }
 
-impl<F: AsPath> FileGroup<F> {
+#[cfg(test)]
+impl<F: AsRef<Path>> FileGroup<F> {
+    fn paths(&self) -> Vec<Path> {
+        self.files.iter().map(|f| f.as_ref().clone()).collect_vec()
+    }
+}
+
+impl<F: AsRef<Path> + AsRef<FileId>> FileGroup<F> {
     /// Returns true if the file group should be forwarded to the next grouping stage,
     /// because the number of duplicate files is higher than the maximum allowed number of replicas.
     ///
@@ -290,7 +302,8 @@ impl<F: AsPath> FileGroup<F> {
                     // fast-path, equivalent to the code in the else branch, but way faster
                     self.file_count().saturating_sub(rf)
                 } else {
-                    let sub_groups = FileSubGroup::group(&self.files, &filter.root_paths);
+                    let sub_groups =
+                        FileSubGroup::group(&self.files, &filter.root_paths, filter.group_by_id);
                     let sub_group_lengths = sub_groups
                         .into_iter()
                         .map(|sg| sg.files.len())
@@ -312,21 +325,19 @@ impl<F: AsPath> FileGroup<F> {
 
     /// The number of subgroups of paths with distinct root prefix.
     fn subgroup_count(&self, filter: &FileGroupFilter) -> usize {
-        if filter.root_paths.is_empty() {
-            // optimisation: the else branch would compute the same number because each file would
-            // be placed in its own subgroup, but the else branch is more complex and likely slower
-            self.files.len()
-        } else {
-            FileSubGroup::group(&self.files, &filter.root_paths).len()
-        }
+        FileSubGroup::group(&self.files, &filter.root_paths, filter.group_by_id).len()
     }
 
     /// Sorts the files by their path names.
     /// If filter requires grouping by roots, then groups are kept together.
     pub fn sort(&mut self, root_paths: &[Path]) {
-        self.files.sort_by(|f1, f2| f1.path().cmp(f2.path()));
+        self.files.sort_by(|f1, f2| {
+            let p1: &Path = f1.as_ref();
+            let p2: &Path = f2.as_ref();
+            p1.cmp(p2)
+        });
         if !root_paths.is_empty() {
-            self.files = FileSubGroup::group(self.files.drain(..), root_paths)
+            self.files = FileSubGroup::group(self.files.drain(..), root_paths, true)
                 .into_iter()
                 .flat_map(|g| g.files)
                 .collect()
@@ -334,6 +345,12 @@ impl<F: AsPath> FileGroup<F> {
     }
 }
 
+impl<T> AsRef<FileGroup<T>> for FileGroup<T> {
+    fn as_ref(&self) -> &FileGroup<T> {
+        self
+    }
+}
+
 /// A subgroup of identical files, typically smaller than a `FileGroup`.
 /// A subgroup is formed by files sharing the same path prefix, e.g. files on the same volume.
 /// In terms of file deduplication activities, a subgroup is an atomic entity -
@@ -350,26 +367,48 @@ impl<F> FileSubGroup<F> {
     pub fn single(f: F) -> FileSubGroup<F> {
         FileSubGroup { files: vec![f] }
     }
+    pub fn push(&mut self, file: F) {
+        self.files.push(file)
+    }
 }
 
-impl<F: AsPath> FileSubGroup<F> {
+impl<F: AsRef<Path> + AsRef<FileId>> FileSubGroup<F> {
     /// Splits a group of files into subgroups.
     ///
     /// Files that share the same prefix found in the roots array are placed in the same subgroup.
     /// The result vector is ordered primarily by the roots, and files having the same root have
-    /// the same order as they came from the input iterator.
-    pub fn group(files: impl IntoIterator<Item = F>, roots: &[Path]) -> Vec<FileSubGroup<F>> {
-        let mut sub_groups = Vec::from_iter(roots.iter().map(|_| FileSubGroup::empty()));
+    /// the same order as they came from the input iterator. Files with paths that don't start
+    /// with any of the root prefixes are placed last in the result, in the same order as the input.
+    ///
+    /// If `group_by_id` is set, files with the same `FileId` are also grouped together.
+    /// In this case, the order of groups follows the order of input files, i.e. the input vector
+    /// is scanned and a new group is appended at the end each time a file with a
+    /// distinct id appears in the input.
+    ///
+    /// If both `roots` is not empty and `group_by_id` is set,
+    /// grouping by prefixes takes precedence over grouping by identifiers,
+    /// so a file with the same id can be placed in two different prefix groups.
+    ///
+    pub fn group(
+        files: impl IntoIterator<Item = F>,
+        roots: &[Path],
+        group_by_id: bool,
+    ) -> Vec<FileSubGroup<F>> {
+        let mut prefix_groups = Vec::from_iter(roots.iter().map(|_| FileSubGroup::empty()));
+        let mut id_groups = IndexMap::new(); // important: keep order of insertion
         for f in files {
-            let path = f.path();
+            let path: &Path = f.as_ref();
+            let id: FileId = *f.as_ref();
             let root_idx = roots.iter().position(|r| r.is_prefix_of(path));
             match root_idx {
-                Some(idx) => sub_groups[idx].files.push(f),
-                None => sub_groups.push(FileSubGroup::single(f)),
+                Some(idx) => prefix_groups[idx].files.push(f),
+                None if group_by_id => id_groups.entry(id).or_insert(FileSubGroup::empty()).push(f),
+                None => prefix_groups.push(FileSubGroup::single(f)),
             }
         }
-        sub_groups.retain(|sg| !sg.files.is_empty());
-        sub_groups
+        prefix_groups.extend(id_groups.into_values());
+        prefix_groups.retain(|sg| !sg.files.is_empty());
+        prefix_groups
     }
 }
 
@@ -637,9 +676,8 @@ fn group_by_size(ctx: &GroupCtx<'_>, files: Vec<Vec<FileInfo>>) -> Vec<FileGroup
     groups
 }
 
-/// Removes duplicate files matching by full-path or by inode-id.
-/// Deduplication by inode-id is not performed if the flag to preserve hard-links (-H) is set.
-fn deduplicate<F>(ctx: &GroupCtx<'_>, files: &mut Vec<FileInfo>, progress: F)
+/// Removes files with duplicate path names.
+fn deduplicate<F>(files: &mut Vec<FileInfo>, progress: F)
 where
     F: Fn(&Path) + Sync + Send,
 {
@@ -651,26 +689,13 @@ where
     for (_, file_group) in groups.into_iter() {
         if file_group.len() == 1 {
             files.extend(file_group.into_iter().inspect(|p| progress(&p.path)));
-        } else if ctx.config.hard_links || ctx.config.symbolic_links {
+        } else {
             files.extend(
                 file_group
                     .into_iter()
                     .inspect(|p| progress(&p.path))
                     .unique_by(|p| p.path.hash128()),
             )
-        } else {
-            // The BTreeMap inside GroupMap results in a random ordering, so repeated runs
-            // will list a different representative filename first for a given inode.
-            // By ordering these the output becomes deterministic.
-            let mut file_group = file_group;
-            file_group.sort_by(|f1, f2| f1.path.cmp(&f2.path));
-
-            files.extend(
-                file_group
-                    .into_iter()
-                    .inspect(|p| progress(&p.path))
-                    .unique_by(|p| file_id_or_log_err(&p.path, ctx.log)),
-            )
         }
     }
 }
@@ -685,7 +710,7 @@ fn remove_same_files(
 
     let groups: Vec<_> = groups
         .into_par_iter()
-        .update(|g| deduplicate(ctx, &mut g.files, |_| progress.tick()))
+        .update(|g| deduplicate(&mut g.files, |_| progress.tick()))
         .filter(|g| g.matches(&ctx.group_filter))
         .collect();
 
@@ -694,7 +719,7 @@ fn remove_same_files(
         "Found {} ({}) candidates after grouping by paths {}",
         stats.0,
         stats.1,
-        if ctx.config.hard_links {
+        if ctx.config.match_links {
             ""
         } else {
             "and file identifiers"
@@ -752,8 +777,7 @@ fn handle_fetch_physical_location_err(
     const MAX_ERR_COUNT_TO_LOG: u32 = 10;
     let device = &ctx.devices[file_info.get_device_index()];
     let counter = &err_counters[device.index];
-    if error.kind() == io::ErrorKind::Unsupported || error.raw_os_error() == Some(libc::EOPNOTSUPP)
-    {
+    if crate::error::error_kind(&error) == io::ErrorKind::Unsupported {
         if counter.swap(MAX_ERR_COUNT_TO_LOG, Ordering::Release) < MAX_ERR_COUNT_TO_LOG {
             ctx.log.warn(format!(
                 "File system {} on device {} doesn't support FIEMAP ioctl API. \
@@ -1033,7 +1057,7 @@ fn group_by_contents(
 /// // print standard fclones report to stdout:
 /// write_report(&config, &log, &groups).unwrap();
 /// ```
-pub fn group_files(config: &GroupConfig, log: &Log) -> Result<Vec<FileGroup<Path>>, Error> {
+pub fn group_files(config: &GroupConfig, log: &Log) -> Result<Vec<FileGroup<FileInfo>>, Error> {
     let spinner = log.spinner("Initializing");
     let ctx = GroupCtx::new(config, log)?;
 
@@ -1043,7 +1067,7 @@ pub fn group_files(config: &GroupConfig, log: &Log) -> Result<Vec<FileGroup<Path
     let mut size_groups_pruned = remove_same_files(&ctx, size_groups);
     update_file_locations(&ctx, &mut size_groups_pruned);
 
-    let groups = match &ctx.transform {
+    let mut groups = match &ctx.transform {
         Some(transform) => group_transformed(&ctx, transform, size_groups_pruned),
         _ => {
             let prefix_len = prefix_len(&ctx.devices, flat_iter(&size_groups_pruned));
@@ -1052,15 +1076,6 @@ pub fn group_files(config: &GroupConfig, log: &Log) -> Result<Vec<FileGroup<Path
             group_by_contents(&ctx, prefix_len, suffix_groups)
         }
     };
-    let mut groups: Vec<_> = groups
-        .into_par_iter()
-        .map(|g| FileGroup {
-            file_len: g.file_len,
-            file_hash: g.file_hash,
-            files: g.files.into_iter().map(|fi| fi.path).collect(),
-        })
-        .collect();
-    groups.retain(|g| g.files.len() < ctx.config.rf_under());
     groups.par_sort_by_key(|g| Reverse((g.file_len, g.file_hash)));
     groups
         .par_iter_mut()
@@ -1078,7 +1093,11 @@ pub fn group_files(config: &GroupConfig, log: &Log) -> Result<Vec<FileGroup<Path
 ///
 /// # Errors
 /// Returns [`io::Error`] on I/O write error or if the output file cannot be created.
-pub fn write_report(config: &GroupConfig, log: &Log, groups: &[FileGroup<Path>]) -> io::Result<()> {
+pub fn write_report(
+    config: &GroupConfig,
+    log: &Log,
+    groups: &[FileGroup<FileInfo>],
+) -> io::Result<()> {
     let now = Local::now();
 
     let total_count = file_count(groups.iter());
@@ -1172,11 +1191,19 @@ mod test {
             file_hash: FileHash(0),
             files: vec![
                 FileInfo {
+                    id: FileId {
+                        device: 1,
+                        inode: 1,
+                    },
                     len: FileLen(200),
                     location: 0,
                     path: Path::from("file1"),
                 },
                 FileInfo {
+                    id: FileId {
+                        device: 1,
+                        inode: 2,
+                    },
                     len: FileLen(200),
                     location: 35847587,
                     path: Path::from("file2"),
@@ -1208,6 +1235,10 @@ mod test {
                 file_len: FileLen(200),
                 file_hash: FileHash(0),
                 files: vec![FileInfo {
+                    id: FileId {
+                        device: 1,
+                        inode: 1,
+                    },
                     len: FileLen(200),
                     location: 0,
                     path: Path::from("file1"),
@@ -1217,6 +1248,10 @@ mod test {
                 file_len: FileLen(500),
                 file_hash: FileHash(0),
                 files: vec![FileInfo {
+                    id: FileId {
+                        device: 1,
+                        inode: 2,
+                    },
                     len: FileLen(200),
                     location: 35847587,
                     path: Path::from("file2"),
@@ -1244,6 +1279,10 @@ mod test {
             file_len: FileLen(200),
             file_hash: FileHash(0),
             files: vec![FileInfo {
+                id: FileId {
+                    device: 1,
+                    inode: 1,
+                },
                 len: FileLen(200),
                 location: 0,
                 path: Path::from("file1"),
@@ -1275,11 +1314,19 @@ mod test {
             file_hash: FileHash(0),
             files: vec![
                 FileInfo {
+                    id: FileId {
+                        device: 1,
+                        inode: 1,
+                    },
                     len: FileLen(200),
                     location: 0,
                     path: Path::from("file1"),
                 },
                 FileInfo {
+                    id: FileId {
+                        device: 1,
+                        inode: 2,
+                    },
                     len: FileLen(200),
                     location: 35847587,
                     path: Path::from("file2"),
@@ -1310,6 +1357,10 @@ mod test {
                 file_len: FileLen(0),
                 file_hash: FileHash(0),
                 files: vec![FileInfo {
+                    id: FileId {
+                        device: 1,
+                        inode: i as InodeId,
+                    },
                     len: FileLen(0),
                     location: i as u64,
                     path: Path::from(format!("file{}", i)),
@@ -1397,8 +1448,8 @@ mod test {
 
             let results = group_files(&config, &log).unwrap();
             assert_eq!(results.len(), 2);
-            assert_eq!(results[0].files, vec![Path::from(file1.canonicalize())]);
-            assert_eq!(results[1].files, vec![Path::from(file2.canonicalize())]);
+            assert_eq!(results[0].paths(), vec![Path::from(file1.canonicalize())]);
+            assert_eq!(results[1].paths(), vec![Path::from(file2.canonicalize())]);
         });
     }
 
@@ -1475,13 +1526,14 @@ mod test {
             hard_link(&file1, &file2).unwrap();
 
             let log = test_log();
+
             let mut config = GroupConfig::default();
             config.paths = vec![file1.into(), file2.into()];
-            config.unique = true;
+            config.unique = true; // hardlinks to a common file should be treated as one file
 
             let results = group_files(&config, &log).unwrap();
             assert_eq!(results.len(), 1);
-            assert_eq!(results[0].files.len(), 1);
+            assert_eq!(results[0].files.len(), 2);
         });
     }
 
@@ -1496,16 +1548,31 @@ mod test {
 
             let log = test_log();
             let mut config = GroupConfig::default();
+
+            // If both hard_links and symbolic_links is set to true, symbolic links should
+            // be treated as duplicates.
             config.paths = vec![file1.into(), file2.into()];
+            config.match_links = true;
             config.symbolic_links = true;
 
             let results = group_files(&config, &log).unwrap();
             assert_eq!(results.len(), 1);
             assert_eq!(results[0].files.len(), 2);
 
+            // Symbolic links should be totally ignored:
             config.symbolic_links = false;
             let results = group_files(&config, &log).unwrap();
             assert_eq!(results.len(), 0);
+
+            // If hard_links is set to false and symbolic_links to true,
+            // a symlink to a file should be reported, but not treated as a duplicate:
+            config.unique = true;
+            config.symbolic_links = true;
+            config.match_links = false;
+
+            let results = group_files(&config, &log).unwrap();
+            assert_eq!(results.len(), 1);
+            assert_eq!(results[0].files.len(), 2);
         });
     }
 
@@ -1519,7 +1586,7 @@ mod test {
             let file1 = Path::from(file1);
             config.paths = vec![file1.clone(), file1.clone(), file1.clone()];
             config.unique = true;
-            config.hard_links = true;
+            config.match_links = true;
 
             let results = group_files(&config, &log).unwrap();
             assert_eq!(results.len(), 1);
@@ -1544,7 +1611,7 @@ mod test {
             let mut config = GroupConfig::default();
             config.paths = vec![file1.into(), file2.into()];
             config.unique = true;
-            config.hard_links = true;
+            config.match_links = true;
 
             let results = group_files(&config, &log).unwrap();
             assert_eq!(results.len(), 1);
@@ -1609,36 +1676,40 @@ mod test {
 
     #[test]
     fn split_to_subgroups() {
+        fn file(path: &str, id: InodeId) -> FileInfo {
+            FileInfo {
+                path: Path::from(path),
+                id: FileId {
+                    inode: id,
+                    device: 0,
+                },
+                len: FileLen(1024),
+                location: id * 1024,
+            }
+        }
+
         let roots = vec![Path::from("/r0"), Path::from("/r1"), Path::from("/r2")];
         let files = vec![
-            Path::from("/r1/f1a"),
-            Path::from("/r2/f2a"),
-            Path::from("/r2/f2b"),
-            Path::from("/r1/f1b"),
-            Path::from("/r1/f1c"),
-            Path::from("/r3/f3a"),
-            Path::from("/r2/f2c"),
+            file("/r1/f1a", 0),
+            file("/r2/f2a", 1),
+            file("/r2/f2b", 2),
+            file("/r1/f1b", 3),
+            file("/r1/f1c", 4),
+            file("/r3/f3a", 5),
+            file("/r2/f2c", 6),
         ];
-        let groups = FileSubGroup::group(files, &roots);
+        let groups = FileSubGroup::group(files, &roots, true);
         assert_eq!(
             groups,
             vec![
                 FileSubGroup {
-                    files: vec![
-                        Path::from("/r1/f1a"),
-                        Path::from("/r1/f1b"),
-                        Path::from("/r1/f1c"),
-                    ]
+                    files: vec![file("/r1/f1a", 0), file("/r1/f1b", 3), file("/r1/f1c", 4),]
                 },
                 FileSubGroup {
-                    files: vec![
-                        Path::from("/r2/f2a"),
-                        Path::from("/r2/f2b"),
-                        Path::from("/r2/f2c")
-                    ]
+                    files: vec![file("/r2/f2a", 1), file("/r2/f2b", 2), file("/r2/f2c", 6)]
                 },
                 FileSubGroup {
-                    files: vec![Path::from("/r3/f3a")]
+                    files: vec![file("/r3/f3a", 5)]
                 }
             ]
         )
